import pandas as pd
from google.cloud import storage
from pathlib import Path


def get_csv_from_gcs(bucket_name, path):
    full_path = 'gs://{bucket}/{path}'.format(bucket=bucket_name, path=path)
    file_type = path.split(".")[-1]
    if file_type == 'csv':
        result_df = pd.read_csv(full_path)
    elif file_type in ('xlsx','xls'):
        with open(full_path, 'rb') as f: # For XLSX file, read in binary format to prevent decode errors
            result_df = pd.read_excel(f, engine='openpyxl', skiprows = [0]) # Ignore first row header from source data
    return result_df


def get_all_file_from_gcs_folder(bucket_name, path, file_type='csv'):
    # Get all file with specific type (XLS or CSV) and return combined dataframe
    file_dict = {}
    try:
        storage_client = storage.Client()
        bucket = storage_client.get_bucket(bucket_name)
        for blob in bucket.list_blobs(prefix=str(path)):
            current_path = Path(blob.name)
            if str(current_path) != str(path):
                full_path = 'gs://{bucket}/{path}'.format(bucket=bucket_name, path=str(current_path))
                if file_type == 'csv':
                    file = pd.read_csv(full_path)
                elif file_type == 'excel':
                    with open(full_path, 'rb') as f: # For XLSX file, read in binary format to prevent decode errors
                        file = pd.read_excel(f, engine='openpyxl', skiprows = [0]) # Ignore first row header from source data
                file_dict[current_path] = file
    except Exception as e:
        raise e
    return file_dict



from flask import Flask, request, jsonify
from flask_restful import Api, Resource
import pandas as pd
from database.bigquery import util as bq_until
from database.bigquery import bq_sql_parser
from database.bigquery.bq_sql_parser import TableInfo


class TableInfoResource(Resource):
    def get(self, dataset, table_name):
        try:
            schema = bq_until.get_table_info(dataset, table_name)
            return jsonify({"schema": str(schema)})
        except Exception as e:
            return {"error": str(e)}, 500


class CreateTableResource(Resource):
    def post(self):
        data = request.get_json()
        table_name = data.get('table_name')
        column_info_dict = data.get('column_info_dict')
        try:
            bq_until.create_table(table_name, column_info_dict=column_info_dict)
            return {"message": f"Table {table_name} created."}
        except Exception as e:
            return {"error": str(e)}, 500


class CreatePartitionTableResource(Resource):
    def post(self):
        data = request.get_json()
        table_name = data.get('table_name')
        project_id = data.get('project_id')
        dataset = data.get('dataset')
        column_info_dict = data.get('column_info_dict')
        partition_date_trunc = data.get('partition_date_trunc')
        partition_column = data.get('partition_column')
        integer_partition_start = data.get('integer_partition_start', None)
        integer_partition_end = data.get('integer_partition_end', None)
        integer_partition_interval = data.get('integer_partition_interval', None)
        cluster_columns = data.get('cluster_columns', None)
        try:
            bq_until.create_partition_table(project_id, dataset, table_name, column_info_dict, partition_column,
                                            partition_date_trunc, integer_partition_start, integer_partition_end,
                                            integer_partition_interval, cluster_columns)
            return {"message": f"Table {table_name} created."}
        except Exception as e:
            return {"error": str(e)}, 500


class InsertDataResource(Resource):
    def post(self):
        # 檢查是否有檔案上傳
        dataset_id = request.form.get('dataset_id')
        table_id = request.form.get('table_id')
        if 'file' not in request.files:
            return {"error": "No file provided"}, 400

        file = request.files['file']
        if file.filename == '':
            return {"error": "No file selected"}, 400

        # 根據檔案類型讀取資料
        if file.filename.endswith('.csv'):
            df = pd.read_csv(file)
        elif file.filename.endswith('.xlsx'):
            df = pd.read_excel(file, engine='openpyxl')
        else:
            return {"error": "Invalid file type"}, 400
        try:
            validated_df = bq_until.validate_dataframe_type_with_bq(df, dataset_id, table_id)
            bq_until.save_data_to_bq(validated_df, dataset_id, table_id)
        except Exception as e:
            return {"error": e}, 400

        return {"message": "Data uploaded successfully"}, 200


class MergeDataResource(Resource):
    def post(self):
        # 檢查是否有檔案上傳
        project_id = request.form.get('project_id')
        target_table = request.form.get('target_table')
        target_dataset = request.form.get('target_dataset')
        unique_keys = request.form.get('unique_keys')
        unique_keys = unique_keys.split(",")
        update_cols = request.form.get('update_cols', None)
        exclude_update_col = request.form.get('exclude_update_col', None)
        if 'file' not in request.files:
            return {"error": "No file provided"}, 400

        file = request.files['file']
        if file.filename == '':
            return {"error": "No file selected"}, 400

        # 根據檔案類型讀取資料
        if file.filename.endswith('.csv'):
            source_df = pd.read_csv(file)
        elif file.filename.endswith('.xlsx'):
            source_df = pd.read_excel(file, engine='openpyxl')
        else:
            return {"error": "Invalid file type"}, 400
        try:
            if (update_cols and not exclude_update_col) or (update_cols and exclude_update_col):
                bq_until.merge_table(project_id, target_table, target_dataset, source_df, unique_keys,
                                     update_cols)
            else:
                bq_until.merge_table(project_id, target_table, target_dataset, source_df, unique_keys,
                                     exclude_update_col)
        except Exception as e:
            return {"error": e}, 400

        return {"message": "Data uploaded successfully"}, 200


class BQSQLGeneratorResource(Resource):
    def post(self):
        try:
            data = request.get_json()
            table_info_dict = data.get('table_info_dict', None)
            sql = data.get('sql')
            selected_columns = data.get('selected_columns', None)
            partition_col = data.get('partition_col', None)
            partition_start = data.get('partition_start', None)
            partition_end = data.get('partition_end', None)
            if table_info_dict or table_info_dict != {}:
                table_info_map = {table_name: TableInfo(table_info['unique_keys'], table_info['date_time_col'],
                                                        table_info['cast_col_map'])
                                  for table_name, table_info in table_info_dict.items()}
            else:
                raise "table_info_dict is None or empty dict"
            format_sql = bq_sql_parser.generate_sql_for_bq(sql, table_info_map, selected_columns, partition_col,
                                                           partition_start, partition_end)
            return format_sql
        except Exception as e:
            return {"error": str(e)}, 500



import re
from typing import NamedTuple, Dict, List


class TableInfo(NamedTuple):
    selected_columns: List[str]
    result_columns: List[str] = []
    groupby_columns: List[str] = []
    datetime_filter: str = ""
    partition_sql: str = ""

class PartitionInfo(NamedTuple):
    unique_keys: List
    date_time_col: str
    cast_col_map: Dict[str, str] = {}
    partition_col: str = ""
    partition_start: str = ""
    partition_end: str = ""
    datetime_filter: str = ""


def __create_cast_col_sql(column_name, cast_type):
    result = f"CAST({column_name} as {cast_type})"
    return result


def generate_sql_for_bq(
    sql,
    table_info_dict: Dict[str, TableInfo],
):
    # TODO : fix partition col and select columns, have to use partition by table
    format_sql = """
        SELECT * FROM (
            SELECT
                {selected_columns},
                ROW_NUMBER()
                    OVER (PARTITION BY {unique_keys} order by {date_time_col} desc)
                    as row_number
            FROM {table}
            WHERE 
                1 = 1
                {partition_sql}
            )
            WHERE row_number = 1
    """

    def __replace_sql(match, table_info_dict, parse_head, format_sql_tmp):
        table_name = match.group(1)
        only_table_name = table_name.split(".")[-1]
        if only_table_name in table_info_dict:
            table_info = table_info_dict[only_table_name]
            partition_info = PARTITION_MAPPING_INFO[only_table_name]
            unique_keys = partition_info.unique_keys
            selected_columns = table_info.selected_columns
            selected_columns_str = ",".join(selected_columns)
            partition_sql = table_info.partition_sql
            # partition_col = partition_info.partition_col
            # partition_start = partition_info.partition_start
            # partition_end = partition_info.partition_end
            # if partition_col and partition_start and partition_end:
            #     partition_sql = (
            #         f"AND {partition_col} between {partition_start} and {partition_end}"
            #     )
            # else:
            #     partition_sql = ""
            cast_col_map = partition_info.cast_col_map
            unique_keys = [
                __create_cast_col_sql(uk, cast_col_map[uk])
                if uk in cast_col_map
                else uk
                for uk in unique_keys
            ]
            # 根据字典进行替换
            return parse_head.format(
                format_sql_tmp.format(
                    unique_keys=",".join(unique_keys),
                    date_time_col=partition_info.date_time_col,
                    table=table_name,
                    selected_columns=selected_columns_str,
                    partition_sql=partition_sql
                )
            )
        else:
            parse_head = parse_head.replace("({0})", "{0}")
            return parse_head.format(table_name)

    from_pattern = r"(?i)\bFROM\s+([^\s()]+)"
    join_pattern = r"(?i)\bJOIN\s+([^\s()]+)"
    format_sql_tmp = format_sql
    for parse_name, pattern in zip(["FROM", "JOIN"], [from_pattern, join_pattern]):
        parse_head = r"{0} ".format(parse_name) + "({0})"
        sql = re.sub(
            pattern,
            lambda match: __replace_sql(
                match, table_info_dict, parse_head, format_sql_tmp
            ),
            sql,
        )
    return sql


def create_sql_from_basic_sql(sql_input: str, table_info_dict: dict) -> str:
    # Get First Table Info as main table
    table_name = list(table_info_dict.keys())[0]

    # Get table info values
    table_info = table_info_dict[table_name]

    result_columns = table_info.result_columns
    result_columns_str = ",".join(result_columns)

    groupby_columns = table_info.groupby_columns
    groupby_columns_str = ",".join(groupby_columns)
    groupby_sql = f"GROUP BY {groupby_columns_str}"

    if datetime_filter := table_info.datetime_filter:
        datetime_filter_str = f"AND {datetime_filter}"
    else:
        datetime_filter_str = ""

    # Create basic SQL
    basic_sql = sql_input.format(
        result_columns=result_columns_str,
        groupby_sql=groupby_sql,
        datetime_filter=datetime_filter_str,
    )

    # basic SQL as input for generating final SQL
    result_sql = generate_sql_for_bq(basic_sql, table_info_dict)
    return result_sql



PARTITION_MAPPING_INFO = {
    "sales_order_level": PartitionInfo(
        unique_keys=["sale_id", "merchant_id"],
        date_time_col="bq_create_time",
    ),
    "sales_prod_level": PartitionInfo(
        unique_keys=["sale_id", "sale_price", "prod_id", "shop_id"],
        date_time_col="bq_create_time",
        cast_col_map={"sale_price": "string"},
    ),
    "mapping_table": PartitionInfo(
        unique_keys=["keys"],
        date_time_col="bq_create_time",
    ),
    "payment_info": PartitionInfo(
        unique_keys=["sale_id", "payment_method", "sale_sno"],
        date_time_col="bq_create_time",
    ),
    "material_consumption_record": PartitionInfo(
        unique_keys=["merchant_id", "shop_id", "material"],
        date_time_col="order_time",
    ),
    "sales_forecast": PartitionInfo(
        unique_keys=["merchant_id", "shop_id", "prod_id"],
        date_time_col="order_time",
    )
}



import time
from typing import List, Dict

import pandas as pd
from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from datetime import datetime

def get_table_info(dataset, table_name, credentials=None):
    client = bigquery.Client(credentials=credentials)
    # table_id = 'your-project.your_dataset.your_table'
    full_table_id = f'{dataset}.{table_name}'
    table = client.get_table(full_table_id)
    return table.schema

def is_valid_datetime(date_value):
# Validate if datetime value is in format %Y-%m-%d, if not then return False
  try:
    date_value = datetime.strptime(date_value, "%Y-%m-%d")  # Example format, adjust as needed
    return True
  except Exception:
    return False

def validate_dataframe_type_with_bq(df, target_dataset, target_table, credentials=None):
    try:
        table_schema = get_table_info(target_dataset, target_table, credentials=credentials)
        for col_info in table_schema:
            if col_info.name in df.columns:
                if col_info.field_type == 'STRING':
                    df[col_info.name] = df[col_info.name].astype(str)
                    df.loc[df[col_info.name] == 'None', col_info.name] = None
                elif col_info.field_type == 'FLOAT':
                    df[col_info.name] = df[col_info.name].astype(float)
                elif col_info.field_type == 'INTEGER':
                    df[col_info.name] = df[col_info.name].astype('Int64')
                elif col_info.field_type == 'TIMESTAMP':
                    df[col_info.name] = pd.to_datetime(df[col_info.name])
                elif col_info.field_type == 'DATETIME':
                    df[col_info.name] = pd.to_datetime(df[col_info.name]).dt.tz_localize(None)
                else:
                    df[col_info.name] = df[col_info.name].astype(object)
                    df.loc[df[col_info.name] == 'None', col_info.name] = None
    except Exception as e:
        print("[ERROR] --> Convert column type error,or Key Error")
        raise e
    return df


def save_data_to_bq(insert_df, dataset_id, table_id, credentials=None, schema = None):
    try:
        client = bigquery.Client(credentials=credentials)
        # 資料表參數
        print("Dataset_id: {}  table_id: {}".format(dataset_id, table_id))
        table_ref = client.dataset(dataset_id).table(table_id)
        print("Create table ref done.")
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = False # Avoid auto-detection, to make sure use provided schema to match
        #job_config.source_format = bigquery.SourceFormat.CSV
        if schema: # If schema provided, may be creating _tmp table for merging
            job_config.schema = schema
            print("received provided schema!")
        else: # if schema not provided, then use table name to get the schema
            table_schema = get_table_info(dataset_id, table_id, credentials=credentials)
            job_config.schema = table_schema # manually define table_schema from BQ table
        
        print("Received schema from bigquery: {}".format(job_config.schema))
        job = client.load_table_from_dataframe(insert_df, table_ref, job_config=job_config)  # Very large scale dataframe, but cannot ignore missing columns
        job.result()
        print("Done insertion.")

    except Exception as e:
        raise e


def query_multiple_data_from_bq(sql_list, data_name, credentials=None):
    try:
        client = bigquery.Client(credentials=credentials)
        result_dict = {}
        for sql, alias in zip(sql_list, data_name):
            results = client.query(sql)
            df = results.to_dataframe()
            result_dict[alias] = df
    except Exception as e:
        raise e
    return result_dict


def check_table_exist(table_id, credentials=None):
    client = bigquery.Client(credentials=credentials)

    try:
        client.get_table(table_id)  # Make an API request.
        print("Table {} already exists.".format(table_id))
        return True
    except NotFound:
        print("Table {} is not found.".format(table_id))
        return False


def create_table(table_name, column_info_dict: Dict[str, Dict[str, str]]):
    client = bigquery.Client()
    schema = [bigquery.SchemaField(col, val['type'], mode=val.get("mode", None))
              for col, val in column_info_dict.items()]
    table = bigquery.Table(table_name, schema=schema)
    table = client.create_table(table)  # Make an API request.
    print(
        "Created table {}.{}.{}".format(table.project, table.dataset_id, table.table_id)
    )


def create_partition_table(project_id, dataset, table_name, column_info_dict: Dict[str, Dict[str, str]],
                           partition_column: str,
                           partition_date_trunc: str, integer_partition_start: int = None,
                           integer_partition_end: int = None, integer_partition_interval: int = None,
                           cluster_columns: List[str] = None):
    try:
        client = bigquery.Client()
        schema = [bigquery.SchemaField(col, val['type'], mode=val.get("mode", None))
                  for col, val in column_info_dict.items()]
        full_table_name = f"{project_id}.{dataset}.{table_name}"
        table = bigquery.Table(full_table_name, schema=schema)
        if cluster_columns:
            table.clustering_fields = cluster_columns
        integer_partition = False
        if partition_date_trunc.upper() == 'DAY':
            date_trunc_type = bigquery.TimePartitioningType.DAY
        elif partition_date_trunc.upper() == 'MONTH':
            date_trunc_type = bigquery.TimePartitioningType.MONTH
        elif partition_date_trunc.upper() == 'HOUR':
            date_trunc_type = bigquery.TimePartitioningType.HOUR
        elif partition_date_trunc.upper() == 'YEAR':
            date_trunc_type = bigquery.TimePartitioningType.YEAR
        elif partition_date_trunc.upper() == 'INTEGER':
            integer_partition = True
        else:
            raise "not match any partition date_trunc"
        if integer_partition:
            table.range_partitioning = bigquery.RangePartitioning(
                # To use integer range partitioning, select a top-level REQUIRED /
                # NULLABLE column with INTEGER / INT64 data type.
                field=partition_column,
                range_=bigquery.PartitionRange(start=integer_partition_start, end=integer_partition_end,
                                               interval=integer_partition_interval),
            )
        else:
            table.time_partitioning = bigquery.TimePartitioning(
                type_=date_trunc_type,
                field=partition_column,  # name of column to use for partitioning
            )
        table = client.create_table(table)  # Make an API request.

        print(
            "Created table {}.{}.{}".format(table.project, table.dataset_id, table.table_id)
        )
    except Exception as e:
        raise e


def merge_table(project_name: str, target_table: str, target_dataset, source_df: pd.DataFrame, unique_keys: List[str],
                update_cols: List[str] = None, exclude_update_col: List[str] = None, credentials=None):
    # Merging: use source_df to add into target table
    # if sale_id, shop_id already exist, then update value. Else insert new row from source_df into target table
    try:
        print("Start Merge Table")
        client = bigquery.Client(credentials=credentials)
        merge_sql = """MERGE `{project_name}.{target_dataset}.{target_table}` T
                        USING `{project_name}.{target_dataset}.{target_table}_tmp` S
                        ON {merge_condition}
                        WHEN MATCHED THEN
                        update set {match_sql}
                        WHEN NOT MATCHED THEN
                        INSERT ({T_columns}) values ({S_columns})"""
        merge_condition = " and ".join([f"T.{col} = S.{col}" for col in unique_keys])
        if update_cols:
            match_sql = ",".join([f"T.{col} = S.{col}" for col in update_cols])
        elif exclude_update_col:
            match_sql = ",".join([f"T.{col} = S.{col}" for col in source_df.columns
                                  if col not in exclude_update_col or col not in unique_keys])
        else:
            match_sql = ",".join([f"T.{col} = S.{col}" for col in source_df.columns if col.lower() not in unique_keys])

       
        S_columns = ",".join([f"S.{col}" for col in source_df.columns])
        T_columns = S_columns.replace('S.', '')


        merge_sql = merge_sql.format(target_dataset=target_dataset, target_table=target_table,
                                     merge_condition=merge_condition, match_sql=match_sql, project_name=project_name,
                                     T_columns = T_columns, S_columns = S_columns)
        print(f"Merge SQL : {merge_sql}")
        validated_df = validate_dataframe_type_with_bq(source_df, target_dataset, target_table)
        target_schema = get_table_info(target_dataset, target_table)
        print(f"Target dataset {target_dataset} target table: {target_table} target schema: {target_schema}")
        
        print("Start create tmp table")
        target_tmp_table = target_table + "_tmp"

        save_data_to_bq(validated_df, target_dataset, target_tmp_table, schema = target_schema) # tmp
        print("create tmp table End")
        print("Start merge")
        start = time.time()
        query_job = client.query(merge_sql)
        query_job.result()
        total_time = str(time.time() - start)
        print(f"End merge : {total_time} seconds")
        client.delete_table(f"{target_dataset}.{target_table}_tmp", not_found_ok=True)
    except Exception as e:
        print(e)
        print("[ERROR]--> Merge Table Error")



import json

from google.cloud import storage
from flask import Response, request
from flask_restful import Resource
import pandas as pd
from database.cloud_storage import util as gcs_util
from pathlib import Path
from model.file_info import FileExtension
from io import BytesIO
import zipfile


class GetFileFromGCS(Resource):
    def post(self):
        try:
            data = request.get_json()
            bucket_name = data.get('bucket_name')
            path = data.get('path')
            file_type = data.get('file_type', None)
            extension = FileExtension(Path(path).suffix.upper())
            if file_type:
                extension = FileExtension(file_type.upper())
            if extension == FileExtension.csv:
                df = gcs_util.get_csv_from_gcs(bucket_name, path)
                return Response(df.to_csv(index=False), mimetype='text/csv',
                                headers={"Content-disposition": "attachment; filename=dataframe.csv"})
            elif extension == FileExtension.xlsx:
                df = gcs_util.get_csv_from_gcs(bucket_name, path)
                output = pd.ExcelWriter('dataframe.xlsx', engine='openpyxl')
                df.to_excel(output, index=False)
                output.save()
                output.seek(0)
                return Response(output.read(),
                                mimetype="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
                                headers={"Content-disposition": "attachment; filename=dataframe.xlsx"})
            elif extension == FileExtension.json:
                json_data = json.loads(gcs_util.get_file_as_string_from_gcs(bucket_name, path))
                return json_data
            else:
                return {"error": "not match any extension"}, 500
        except Exception as e:
            return {"error": str(e)}, 500


class GetWholeFileFromGCSFolder(Resource):
    def post(self):
        try:
            data = request.get_json()
            bucket_name = data.get('bucket_name')
            path = data.get('path')
            client = storage.Client()
            bucket = client.bucket(bucket_name)
            # 壓縮檔案夾中的所有檔案
            memory_file = BytesIO()
            with zipfile.ZipFile(memory_file, 'w') as zf:
                for blob in bucket.list_blobs(prefix=path):
                    file_buffer = BytesIO()
                    blob.download_to_file(file_buffer)
                    file_buffer.seek(0)
                    zf.writestr(blob.name, file_buffer.getvalue())

            # 設定回傳的壓縮檔
            memory_file.seek(0)
            return Response(memory_file, mimetype='application/zip',
                            headers={"Content-disposition": f"attachment; filename=files.zip"})

        except Exception as e:
            return {"error": str(e)}, 500


class UploadFileToGCS(Resource):
    def post(self):
        try:
            bucket_name = request.form.get('bucket_name')
            path = request.form.get('path')
            if 'file' not in request.files:
                return {"error": "No file provided"}, 400

            file = request.files['file']
            if file.filename == '':
                return {"error": "No file selected"}, 400
            gcs_util.upload_file_to_gcs(bucket_name, path, file)
            return "", 200
        except Exception as e:
            return {"error": str(e)}, 500



from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaIoBaseDownload
from googleapiclient.http import MediaIoBaseUpload
from google.oauth2 import service_account
import pandas as pd
import io
from pathlib import Path
from zipfile import ZipFile
import requests

class GDriveToolkit:
    def __init__(self, shared_drive_id, key_file_path, path=None):
        self.SCOPES = ['https://www.googleapis.com/auth/drive']
        self.KEY_FILE_PATH = key_file_path
        self.creds = service_account.Credentials.from_service_account_file(self.KEY_FILE_PATH, scopes=self.SCOPES)
        self.service = build('drive', 'v3', credentials=self.creds)
        self.shared_drive_id = shared_drive_id
        self.path = path
        if self.path:
            path_parts = Path(self.path)
            self.root_list = [path for path in path_parts.parts]
            if len(self.root_list) > 1:
                self.folder_dict = self.__partial_traversal(self.shared_drive_id, self.root_list[1:])
            else:
                self.folder_dict = self.__full_traversal(self.shared_drive_id)
        else:
            self.folder_dict = self.__full_traversal(self.shared_drive_id)

    def __get_path_info(self, path):

        # get the parts of path
        path_parts = Path(path)

        if '.' in path_parts.name:  # Check if the last part of path is a file
            path_parts = path_parts.parent

        # search the corresponding file id based on the path_parts name
        current_dict = self.folder_dict
        folder_name = 'default'
        folder_id = 'default'
        path_id_name_list = {}
        for path_part in path_parts.parts[1:]:

            if len(current_dict[path_part].keys()) > 1:
                folder_id = current_dict[path_part]["id"]
                folder_name = path_part
                current_dict = {key: value for key, value in current_dict[path_part].items() if key != 'id'}


            else:
                folder_id = current_dict[path_part]["id"]
                folder_name = path_part

        path_id_name_list[folder_name] = {'id': folder_id}

        return path_id_name_list

    def __partial_traversal(self, folder_id, path_list):
        if len(path_list) > 1:
            query = f" '{folder_id}' in parents and name='{path_list[0]}' and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
            results = self.service.files().list(q=query, spaces='drive',
                                                fields='files(id, name)',
                                                supportsAllDrives=True,
                                                includeItemsFromAllDrives=True,
                                                driveId=self.shared_drive_id, corpora='drive'
                                                ).execute()
            items = results.get("files", [])
            folder_dict = {
                items[0]['name']: {
                    'id': items[0]['id'],
                    **self.__partial_traversal(items[0]['id'], path_list[1:])
                }
            }
            return folder_dict
        elif len(path_list) == 1 and folder_id != self.shared_drive_id:
            query = f" '{folder_id}' in parents and name='{path_list[0]}' and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
            results = self.service.files().list(q=query, spaces='drive',
                                                fields='files(id, name)',
                                                supportsAllDrives=True,
                                                includeItemsFromAllDrives=True,
                                                driveId=self.shared_drive_id, corpora='drive'
                                                ).execute()
            items = results.get("files", [])
            temp_dict = {}
            for item in items:
                temp_dict = self.__full_traversal(item['id'])
            return {items[0]['name']: {'id': items[0]['id'], **temp_dict}}
        else:
            query = f" name = '{path_list[0]}' and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
            results = self.service.files().list(q=query, spaces='drive',
                                                fields='files(id, name)',
                                                supportsAllDrives=True,
                                                includeItemsFromAllDrives=True,
                                                driveId=self.shared_drive_id, corpora='drive'
                                                ).execute()
            items = results.get("files", [])
            temp_dict = {}
            for item in items:
                temp_dict = self.__full_traversal(item['id'])
            return {items[0]['name']: {'id': items[0]['id'], **temp_dict}}

    def __full_traversal(self, folder_id):

        query = f" '{folder_id}' in parents and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
        results = self.service.files().list(q=query, spaces='drive',
                                            fields='files(id, name)',
                                            supportsAllDrives=True,
                                            includeItemsFromAllDrives=True,
                                            driveId=self.shared_drive_id, corpora='drive'
                                            ).execute()

        items = results.get("files", [])
        folder_dict = {}
        for item in items:
            folder_id = item['id']
            folder_name = item['name']
            subfolder_dict = self.__full_traversal(folder_id)
            if subfolder_dict != {}:
                folder_dict[folder_name] = {'id': folder_id, **subfolder_dict}
            else:
                folder_dict[folder_name] = {'id': folder_id}
        return folder_dict

    def get_folder_files(self, path):
        # input the path and return the file name and file id
        folder_info = self.__get_path_info(path)
        path_parts = Path(path)
        folder_name = path_parts.name
        folder_id = folder_info[folder_name]['id']
        query = f" '{folder_id}' in parents and mimeType != 'application/vnd.google-apps.folder' and trashed = false"
        results = self.service.files().list(q=query, spaces='drive',
                                            fields='files(id, name)',
                                            supportsAllDrives=True,
                                            includeItemsFromAllDrives=True,
                                            driveId=self.shared_drive_id, corpora='drive'
                                            ).execute()
        items = results.get("files", [])
        return folder_id, items


    def read_folder_files(self, folder_id, items, sheet_name=0):
        try:
            dataframes = {}
            for item in items:
                file_name = item['name']
                output = self.__file_loader(file_name, folder_id, sheet_name)

                for key in output.keys():
                    if file_name in key:
                        dataframes.update(output)
            return dataframes
        except HttpError as error:
            raise f'An error occurred:{error}'
        except Exception as error:
            raise f'An error occurred:{error}'

    def read_file(self, path, sheet_name=0) -> pd.DataFrame:

        folder_info = self.__get_path_info(path)

        path_parts = Path(path)

        folder_name = path_parts.parent.name
        file_name = path_parts.name

        folder_id = folder_info[folder_name]['id']

        # if the last part is file, read the file
        # if len(file_name) > 0:
        file = pd.DataFrame()
        dataframes = self.__file_loader(file_name, folder_id, sheet_name)
        for key in dataframes.keys():
            if file_name in key:
                file = dataframes[key]
        return file

    def rpt_format_file_loader(self, file_name, folder_id, sheet_name=0):
        def read_csv_as_string(result_text):
            csv_text = result_text.replace("\ufeff", "")
            csv_text_list = [
                text for text in csv_text.split("\n") if not text.startswith("(")
            ]
            return csv_text_list
        
        def separate_table(csv_text_list):
            text_list = []
            tmp_text = []
            for text in csv_text_list:
                if text not in ['\r', '']: # 切分辨識用的符號
                    tmp_text.append(f"\n{text}")
                else:
                    text_list.append(tmp_text)
                    tmp_text = []
            text_list = [t for t in text_list if t != []]
            return text_list

        def read_rpt_file_object_as_dataframe(text_list):
            df_dict = {}
            for text in text_list:
                if text:
                    try:
                        table_name = text[2].split(" ")[0].strip().lower()
                        text_file_object = io.StringIO("".join(text))
                        df = pd.read_fwf(text_file_object)[1:]
                        df_dict[table_name] = df
                    except Exception as e:
                        print(e)
                        print(text)
            return df_dict

        print(f"start read: {file_name}")
        # try:
        query = f"name = '{file_name}' and '{folder_id}' in parents and trashed = false"
        results = self.service.files().list(q=query, spaces='drive',
                                            fields='files(id, name)',
                                            supportsAllDrives=True,
                                            includeItemsFromAllDrives=True,
                                            driveId=self.shared_drive_id, corpora='drive'
                                            ).execute()

        items = results.get('files', [])

        # try:
        output = {}
        for item in items:
            if item['name'].endswith(('.xlsx', '.xls', '.csv')):
                file_id = item['id']
                request = self.service.files().get_media(fileId=file_id)
                # 2023/10/31 改變讀檔案方式: 用URL+token
                token = self.creds.token
                url = "https://www.googleapis.com/drive/v3/files/" + file_id + "?alt=media"
                res = requests.get(url, headers={"Authorization": "Bearer " + token})
                res.encoding = res.apparent_encoding  # 'ISO-8859-1' 轉成 'Big5'
                csv_text_list = read_csv_as_string(res.text)
                text_list = separate_table(csv_text_list)
                df_dict = read_rpt_file_object_as_dataframe(text_list)
                for table, data in df_dict.items():
                    output[(file_name, file_id, table)] = data
        return output


    def __file_loader(self, file_name, folder_id, sheet_name=0):
        print(f"start read: {file_name}")

        try:

            query = f"name = '{file_name}' and '{folder_id}' in parents and trashed = false"
            results = self.service.files().list(q=query, spaces='drive',
                                                fields='files(id, name)',
                                                supportsAllDrives=True,
                                                includeItemsFromAllDrives=True,
                                                driveId=self.shared_drive_id, corpora='drive'
                                                ).execute()

            items = results.get('files', [])

        except HttpError as error:
            raise f'An error occurred:{error}'

        try:
            output = {}
            # csv, xlsx, xls
            for item in items:
                if item['name'].endswith(('.xlsx', '.xls', '.csv')):
                    file_id = item['id']
                    request = self.service.files().get_media(fileId=file_id)
                    downloaded = io.BytesIO()
                    downloader = MediaIoBaseDownload(downloaded, request)
                    done = False
                    while done is False:
                        status, done = downloader.next_chunk()
                    downloaded.seek(0)
                    try:
                        if item['name'].endswith('.csv'):
                            df = pd.read_csv(downloaded)
                        else:
                            df = pd.read_excel(downloaded, engine='openpyxl', sheet_name=sheet_name)
                    except:

                        self.read_as_xml(downloaded)
                        df = pd.read_excel("tmp.xlsx", engine="openpyxl", sheet_name=sheet_name)

                    output[(file_name, file_id)] = df

            return output

        except HttpError as error:
            raise f'An error occurred:{error}'
        except Exception as e:
            raise f'An error occurred:{e}'

    def read_as_xml(self, data):
        with ZipFile(data, 'r') as ZIP:
            list_files = ZIP.namelist()
            for f in list_files:
                ZIP.extract(f)
            style_xml = ZIP.read('xl/styles.xml')
            new_style_str = str(style_xml).replace('horizontal="left"', '"horizontal="center"')

            with open("xl/styles_new.xml", "w") as binary_file:
                binary_file.write(new_style_str)
            with ZipFile('tmp.xlsx', mode='w') as zf:
                # 加入要壓縮的檔案
                for f in ZIP.namelist():
                    if f != 'xl/styles.xml':
                        zf.write(f)
                zf.write("xl/styles_new.xml")

    def move_file(self, file_name, old_path, new_path):

        # check folders is existed, create it if not existed
        self.__check_folder_exist(new_path)

        old_path_parts = Path(old_path)
        new_path_parts = Path(new_path)

        old_folder_name = old_path_parts.name
        new_folder_name = new_path_parts.name

        old_path_info = self.__get_path_info(old_path)
        new_path_info = self.__get_path_info(new_path)

        old_folder_id = old_path_info[old_folder_name]['id']
        new_folder_id = new_path_info[new_folder_name]['id']

        print(f"start moving file {file_name} to {new_path}")
        try:
            query = f" name = '{file_name}' and '{old_folder_id}' in parents and trashed = false"
            results = self.service.files().list(q=query, fields='files(id, name)',
                                                supportsAllDrives=True,
                                                includeItemsFromAllDrives=True,
                                                driveId=self.shared_drive_id, corpora='drive').execute()
            items = results.get('files', [])

            file_id = items[0]['id']

            print(f"Moving file: {file_name}")

            self.service.files().update(fileId=file_id,
                                        addParents=new_folder_id,
                                        removeParents=old_folder_id,
                                        fields="id,parents", supportsAllDrives=True).execute()
            print(f"successfully moved file {file_name} to {new_path}")
        except HttpError as error:
            raise f'An error occurred:{error}'
        except Exception as e:
            raise f'An error occurred:{e}'

    def save_to_drive(self, df, file_name, path):

        # check folders is existed, create it if not existed
        self.__check_folder_exist(path)

        folder_info = self.__get_path_info(path)

        path_parts = Path(path)

        folder_name = path_parts.name

        folder_id = folder_info[folder_name]['id']

        file_extension = Path(file_name).suffix

        try:

            if file_extension == ".csv":
                # .csv is textual data so use StringIO
                data_buffer = io.BytesIO()
                df.to_csv(data_buffer, index=False)
                mimetype = 'text/csv'
            elif file_extension in [".xlsx", ".xls"]:
                # .xlsx and .xls are binary data so use BytesIO
                data_buffer = io.BytesIO()
                df.to_excel(data_buffer, index=False, engine='openpyxl')
                mimetype = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
            else:
                data_buffer = io.BytesIO()
                df.to_excel(data_buffer, index=False, engine='openpyxl')
                mimetype = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'

        except Exception as e:
            raise f'An error occurred:{e}'

        try:
            # check file if it is existed
            query = f"name = '{file_name}' and '{folder_id}' in parents and trashed = false"
            result = self.service.files().list(q=query, spaces='drive',
                                               fields='files(id, name)',
                                               supportsAllDrives=True,
                                               includeItemsFromAllDrives=True,
                                               driveId=self.shared_drive_id, corpora='drive'
                                               ).execute()
            files = result.get('files', [])

            media = MediaIoBaseUpload(data_buffer, mimetype=mimetype, resumable=True)

            if len(files) > 0:
                file_metadata = {
                    'name': file_name
                }
                # file already exists, update it
                file_id = files[0].get('id')
                print(f"'{file_name}' is existed, update it")
                self.service.files().update(fileId=file_id, body=file_metadata, media_body=media, fields='id',
                                            supportsAllDrives=True).execute()
                print(f"'{file_name}' is updated")
            else:
                file_metadata = {
                    'name': file_name,
                    'parents': [folder_id]
                }
                self.service.files().create(body=file_metadata, media_body=media, fields='id',
                                            supportsAllDrives=True).execute()

                print(f"'{file_name}' is saved")

        except Exception as e:
            raise f'An error occurred:{e}'

    def __check_folder_exist(self, path):

        current_dict = self.folder_dict
        path_parts = Path(path)
        flag = False
        path_id = None

        for path_part in path_parts.parts[1:]:

            if path_part in current_dict:
                if len(current_dict[path_part].keys()) > 1:
                    path_id = current_dict[path_part]['id']
                    current_dict = {key: value for key, value in current_dict[path_part].items() if key != 'id'}

            else:
                flag = True
                parent_id = path_id or current_dict['id']
                file_metadata = {
                    "name": path_part,
                    'mimeType': 'application/vnd.google-apps.folder',
                    'parents': [parent_id]
                }
                try:
                    new_folder = self.service.files().create(body=file_metadata, fields="id",
                                                             supportsAllDrives=True).execute()
                    current_dict[path_part] = {"id": new_folder['id']}
                    path_id = None
                    current_dict = current_dict[path_part]
                    print(f"folder {path_part} is not existed and has been created.")
                except HttpError as error:
                    raise f'An error occurred:{error}'

        if flag:
            self.folder_dict = self.__partial_traversal(self.shared_drive_id, self.root_list[1:])
            print("some folders are not existed and have been created.")






import re

from typing import NamedTuple, Dict, List


class TableInfo(NamedTuple):
    unique_keys: List
    date_time_col: str
    cast_col_map: Dict[str, str] = {}


def __create_cast_col_sql(column_name, cast_type):
    result = f"CAST({column_name} as {cast_type})"
    return result


def replace_table_names(sql, replacement, table_info_dict, selected_columns_str="*", partition_sql=""):
    def __replace_sql(match, table_info_dict, parse_head, replacement_tmp):
        table_name = match.group(1)
        only_table_name = table_name.split(".")[-1]
        if only_table_name in table_info_dict:
            table_info = table_info_dict[only_table_name]
            unique_keys = table_info.unique_keys
            cast_col_map = table_info.cast_col_map
            unique_keys = [__create_cast_col_sql(uk, cast_col_map[uk]) if uk in cast_col_map else uk
                           for uk in unique_keys]
            # 根据字典进行替换
            return parse_head.format(replacement_tmp.format(unique_keys=",".join(unique_keys),
                                                            date_time_col=table_info.date_time_col,
                                                            table=table_name, selected_columns=selected_columns_str,
                                                            partition_sql=partition_sql))
        else:
            parse_head = parse_head.replace("({0})", "{0}")
            return parse_head.format(table_name)

    from_pattern = r"(?i)\bFROM\s+([^\s()]+)"
    join_pattern = r"(?i)\bJOIN\s+([^\s()]+)"
    replacement_tmp = replacement
    for parse_name, pattern in zip(["FROM", 'JOIN'], [from_pattern, join_pattern]):
        parse_head = r"{0} ".format(parse_name) + "({0})"
        sql = re.sub(pattern,
                     lambda match: __replace_sql(match, table_info_dict, parse_head, replacement_tmp), sql)
    return sql


def generate_sql_for_bq(sql, table_info: Dict[str, TableInfo], selected_columns: List[str] = None,
                        partition_col=None, partition_start=None, partition_end=None):
    # TODO : fix partition col and select columns, have to use partition by table
    format_sql = """
    SELECT * FROM (
          SELECT
              {selected_columns},
              ROW_NUMBER()
                  OVER (PARTITION BY {unique_keys} order by {date_time_col} desc)
                  as row_number
          FROM {table}
          {partition_sql}
        )
        WHERE row_number = 1
    """
    selected_columns_str = "*"
    partition_sql = ""
    if selected_columns:
        selected_columns_str = ",".join(selected_columns)
    if partition_col and partition_end and partition_start:
        partition_sql = f"{partition_col} between {partition_start} and {partition_end}"

    result_sql = replace_table_names(sql, format_sql, table_info, selected_columns_str, partition_sql)
    return result_sql



from setuptools import find_packages, setup
import pathlib

# 若Discription.md中有中文 須加上 encoding="utf-8"
HERE = pathlib.Path(__file__).parent.resolve()
requirements = (HERE / "requirements.txt").read_text(encoding="utf8")
EX_INSTALL_REQUIRES = {s.strip().split('==')[0]: s.strip().split('==')[1] if len(s.strip().split('==')) > 1 else ""
                       for s in requirements.split("\n")}
INSTALL_REQUIRES = [s.strip() for s in requirements.split("\n")]
setup(
    name="shaku-database-latest",
    version="1.0.10",
    author="shaku-developer",
    author_email="laurencelin@shaku.com.tw",
    description="Shaku Database util",
    long_description="Only for Shaku Company internal use.",
    long_description_content_type="text/markdown",
    url="https://gitlab.com/shaku-com/data-util",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=INSTALL_REQUIRES,
    python_requires='>=3.6',
    extras_require=EX_INSTALL_REQUIRES
)



import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.sql import text


def query(sql, connection_string):
    engine = create_engine(connection_string)
    s = text(sql)
    with engine.connect() as conn:
        result = conn.execute(s)
    df = pd.DataFrame(result.fetchall(), columns=result.keys())
    return df


def insert_update(sql, data, connection_string):
    sql_data = data.to_dict(orient='records')
    engine = create_engine(connection_string)
    s = text(sql)
    with engine.connect() as conn:
        try:
            conn.execute(s, sql_data)
            conn.commit()
        except Exception as e:
            print(e)
            conn.rollback()


def batch_insert_data(sql_list, data_list, connection_string):
    engine = create_engine(connection_string)
    with engine.connect() as conn:
        try:
            for sql, data in zip(sql_list, data_list):
                sql_data = data.to_dict(orient='records')
                s = text(sql)
                conn.execute(s, sql_data)
            conn.commit()
        except Exception as e:
            print(e)
            conn.rollback()



from typing import List, Dict

import pandas as pd
from google.cloud import storage
from pathlib import Path
from google.oauth2 import service_account

def get_csv_from_gcs(bucket_name, path):
    full_path = 'gs://{bucket}/{path}'.format(bucket=bucket_name, path=path)
    file_type = path.split(".")[-1]
    result_df = pd.DataFrame()
    if file_type == 'csv':
        result_df = pd.read_csv(full_path)
    elif file_type == 'xlsx' or file_type == 'xls':
        result_df = pd.read_excel(full_path, engine='openpyxl')
    return result_df


def get_all_file_from_gcs_folder(bucket_name, path, source_info, file_type='csv'):
    # Download all CSV and Excel file under specific bucket folder as dataframe object
    file_dict = {}
    sheet_name = source_info.sheet_name
    try:
        storage_client = storage.Client()
        bucket = storage_client.get_bucket(bucket_name)
        for blob in bucket.list_blobs(prefix=str(path)):
            current_path = Path(blob.name)
            if str(current_path).endswith('sample.txt'): # ignore sample file
                print("Ignore sample file sample.txt")
                continue

            if str(current_path) != str(path):
                full_path = 'gs://{bucket}/{path}'.format(bucket=bucket_name, path=str(current_path))
                file = pd.DataFrame()
                file_type = full_path.split('.')[-1]
                print("Data file full path: ", full_path)
                # Specify string type column during reading file
                col_type_def = {}
                if len(source_info.string_type_column) > 0: # If having pre-defined string column, setup during reading
                    for col in source_info.string_type_column:
                        col_type_def[col] = str

                print("Read file with defined data type: ")
                print(col_type_def)
                if file_type == 'csv':
                    file = pd.read_csv(full_path, dtype = col_type_def)
                elif file_type in ('xls', 'xlsx', 'xlsm'):
                    # Specify sheet name when parsing XLS file for multiple sheets
                    # Specify different parsing rules that applies on all merchant, respective to different data sources
                    if sheet_name:
                        print("Read from sheet name: ", sheet_name)
                        file = pd.read_excel(full_path, sheet_name=sheet_name, dtype = col_type_def)
                    else:
                        file = pd.read_excel(full_path, dtype = col_type_def)
                    

                else:
                    raise ValueError("File not valid type! Invalid file in path: {}".format(full_path))
                file_dict[current_path] = file
    except Exception as e:
        raise ValueError(e)
    return file_dict


def get_file_as_string_from_gcs(bucket_name, path) -> bytes:
    # Load file content from file in bucket folder as string
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(path)

    # Download the contents of the blob as a string and then parse it using json.loads() method
    return blob.download_as_string(client=None)


def get_all_json_from_gcs(bucket_name, folder_path) -> Dict[str, bytes]:
    # Get all JSON object information from bucket folder path
    # Output: bucker file name, json content
    client = storage.Client()
    bucket = client.bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=folder_path)
    result_dict = {}
    for blob in blobs:
        if blob.name.endswith('.json'):
            result_dict[blob.name] = blob.download_as_string(client=None)
    return result_dict


def upload_json_to_gcs(bucket_name, destination_blob_name, json_data):
    # 初始化GCS客户端
    client = storage.Client()

    bucket = client.get_bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_string(json_data, content_type='application/json')


def delete_file_from_gcs(bucket_name, file_name):
    # 初始化 GCS 客户端
    client = storage.Client()
    bucket = client.get_bucket(bucket_name)
    blob = bucket.blob(file_name)
    blob.delete()
    print(f'Successfully deleted file  {file_name} from {bucket_name}.')


def upload_file_to_gcs(target_bucket_name, folder_path, file):
    # Upload local object to GCS bucket
    bucket_folder_path = 'gs://' + str(target_bucket_name) + '/' +  str(folder_path) # Format GCS target folder URL to upload
    folder_file_suffix = '.' + str(bucket_folder_path).split('.')[-1] # Get suffix: .xlsx, .xls, or other type
    bucket_folder_path_csv = str(bucket_folder_path).replace(folder_file_suffix, '.csv')
    print(f"Bucket folder csv path: {bucket_folder_path_csv}")
    file.to_csv(bucket_folder_path_csv, index = False) # save dataframe object to GCS bucket folder
    return {"public_url": bucket_folder_path}

def move_blob(source_bucket: str, source_blob_file: str, destination_bucket: str, destination_blob_file: str):
    """Moves a blob from one bucket to another with a new name."""
    # source_bucket : "your-bucket-name"
    # source_blob_file: Str = "/path/to/file/under/source_bucket.txt"
    # destination_bucket = "destination-bucket-name"
    # destination_blob_file = "/path/to/file/under/destination_bucket.txt"
    storage_client = storage.Client()

    source_bucket = storage_client.bucket(source_bucket)
    source_blob = source_bucket.get_blob(source_blob_file)
    
    destination_bucket = storage_client.bucket(destination_bucket)
    
        # Optional: set a generation-match precondition to avoid potential race conditions
        # and data corruptions. The request is aborted if the object's
        # generation number does not match your precondition. For a destination
        # object that does not yet exist, set the if_generation_match precondition to 0.
        # If the destination object already exists in your bucket, set instead a
        # generation-match precondition using its generation number.
        # There is also an `if_source_generation_match` parameter, which is not used in this example.
    # if target file already exists in bucket folder, set this to None. 
    # if target file not exist yet, set it to 0
    if source_blob.exists(): # if file already exist in bucket folder
        destination_generation_match_precondition = None # set to None only when target destination exist already
    else:
        destination_generation_match_precondition = 0  

    blob_copy = source_bucket.copy_blob(
            source_blob, destination_bucket, destination_blob_file, if_generation_match=destination_generation_match_precondition,
    )